-- Import global variables

local assert, error, ipairs, next, pairs, select, tonumber, tostring, type, unpack, pcall, xpcall =
      assert, error, ipairs, next, pairs, select, tonumber, tostring, type, unpack, pcall, xpcall
local getmetatable, setmetatable, rawequal, rawget, rawset, getfenv, setfenv, loadstring, debugstack =
      getmetatable, setmetatable, rawequal, rawget, rawset, getfenv, setfenv, loadstring, debugstack
local math, string, table = math, string, table
local find, format, gmatch, gsub, tolower, match, toupper, join, split, trim =
      string.find, string.format, string.gmatch, string.gsub, string.lower, string.match, string.upper, string.join, string.split, string.trim
local concat, insert, maxn, remove, sort = table.concat, table.insert, table.maxn, table.remove, table.sort
local max, min, abs, ceil, floor = math.max, math.min, math.abs, math.ceil, math.floor

local AceLibrary = AceLibrary

local GetTime = GetTime
local IsMouselooking = IsMouselooking
local MouselookStart = MouselookStart
local MouselookStop = MouselookStop
local SpellIsTargeting = SpellIsTargeting
local SetMouselookOverrideBinding = SetMouselookOverrideBinding

-- No global variables after this!

local _G = getfenv()
do
  local function gerr(t, k) error(format("Reference to global %q", k), 2) end
  setfenv(1, setmetatable({}, {__index = gerr, __newindex = gerr}))
end


-- ================================================================================================
-- = Localization                                                                                 =
-- ================================================================================================

local L = AceLibrary("AceLocale-2.2"):new("MouseLook")

L:RegisterTranslations("enUS", function() return {
    ["Toggle mouse-look on/off"] = true,
    ["Turn mouse-look on"] = true,
    ["Turn mouse-look off"] = true,
    ["Auto-run"] = true,
    ["Enter mouse-look mode when you begin auto-running."] = true,
    ["Auto-follow"] = true,
    ["Leave mouse-look mode when you begin auto-following a unit."] = true,
    ["Right-click"] = true,
    ["Enter mouse-look mode when holding down the right mouse button for a short amount of time."] = true,
    ["Right-click delay"] = true,
    ["Set the amount of time you need to hold down the right mouse button before entering mouse-look mode."] = true,
    ["Spell-cast"] = true,
    ["Leave mouse-look mode when you cast a spell that needs a mouse-click to target."] = true,
    ["Left-click override"] = true,
    ["Adjust behaviour of the left mouse-button when you are in mouse-look mode."] = true,
    ["Right-click override"] = true,
    ["Adjust behaviour of the right mouse-button when you are in mouse-look mode."] = true,
    ["Do nothing"] = true,
  }
end)

-- L = setmetatable({}, {__index = function(self, key) return key end})


-- ================================================================================================
-- =  Initialization (addon, libraries, etc.)                                                     =
-- ================================================================================================
local AddonTitle = "MouseLook"
_G.BINDING_HEADER_MOUSELOOK = AddonTitle
_G.BINDING_NAME_MOUSELOOK_TOGGLE = L["Toggle mouse-look on/off"]
_G.BINDING_NAME_MOUSELOOK_ON = L["Turn mouse-look on"]
_G.BINDING_NAME_MOUSELOOK_OFF = L["Turn mouse-look off"]

local mixins = {"AceHook-2.1", "AceEvent-2.0", "AceDB-2.0", "AceConsole-2.0", "FuBarPlugin-2.0"}
local hasAceDebug = AceLibrary:HasInstance("AceDebug-2.0")
if hasAceDebug then insert(mixins, "AceDebug-2.0") end

local SE_Movement = AceLibrary("SpecialEvents-Movement-2.0")

local Core = AceLibrary("AceAddon-2.0"):new(unpack(mixins))
Core:RegisterDB("MouseLookDB")

local function noop() end
local dbg, debug, ldebug
if hasAceDebug then
  function dbg() return Core:IsDebugging() end
  function debug(...) return Core:Debug(...) end
  function ldebug(level, ...) return Core:LevelDebug(level, ...) end
  -- Core:SetDebugging(true)
  -- Core:SetDebugLevel(1)
else
  dbg, debug, ldebug = noop, noop, noop
end

function Core:ToString() return AddonTitle end
function Core:GetTitle() return AddonTitle end

local MinRightClickDelay = .1
local MaxRightClickDelay = 1
local DefaultRightClickDelay = .7

local function FixDelay(delay)
  return delay and max(MinRightClickDelay, min(MaxRightClickDelay, floor(delay * 100 + .5) / 100)) or DefaultRightClickDelay
end

local ClickOverrides = {}
do
  ClickOverrides.NONE = L["Do nothing"]
  local bindings = {
    "MOUSELOOK_OFF", "STRAFELEFT", "STRAFERIGHT", "MOVEFORWARD", "MOVEBACKWARD",
    "JUMP", "SITORSTAND", "TOGGLEAUTORUN", "TOGGLERUN",
  }
  for _, binding in ipairs(bindings) do
    ClickOverrides[binding] = _G["BINDING_NAME_"..binding] or binding
  end
end

local function FixClickOverride(click_override)
  if click_override and ClickOverrides[click_override] then return click_override end
  return "NONE"
end


local options
function Core:OnInitialize()
  self:RegisterChatCommand("/mouselook", "/ml", options)

  local defaults = {
    autorun = true,
    autofollow = true,
    rightclick = true,
    rightclick_delay = DefaultRightClickDelay,
    leftclick_override = "NONE",
    rightclick_override = "MOUSELOOK_OFF",
    spellcast = true,
  }

  self:RegisterDefaults('profile', defaults)
end

function Core:OnEnable()
  self:OnProfileEnable()
end

function Core:OnDisable()
  self:MouselookOff()
end

function Core:IsAutoRun()
  return not not self.db.profile.autorun
end

function Core:SetAutoRun(enable)
  self.db.profile.autorun = not not enable
  self:EnableHooksAndEvents()
end

function Core:IsAutoFollow()
  return not not self.db.profile.autofollow
end

function Core:SetAutoFollow(enable)
  self.db.profile.autofollow = not not enable
  self:EnableHooksAndEvents()
end

function Core:IsRightClick()
  return not not self.db.profile.rightclick
end

function Core:SetRightClick(enable)
  self.db.profile.rightclick = not not enable
  self:EnableHooksAndEvents()
end

function Core:GetRightClickDelay()
  return self.db.profile.rightclick_delay
end

function Core:SetRightClickDelay(delay)
  self.db.profile.rightclick_delay = FixDelay(delay)
end

function Core:GetLeftClickOverride()
  return self.db.profile.leftclick_override
end

function Core:SetLeftClickOverride(key)
  self.db.profile.leftclick_override = key
  self:EnableHooksAndEvents()
end

function Core:GetRightClickOverride()
  return self.db.profile.rightclick_override
end

function Core:SetRightClickOverride(key)
  self.db.profile.rightclick_override = key
  self:EnableHooksAndEvents()
end

function Core:IsSpellCast()
  return not not self.db.profile.spellcast
end

function Core:SetSpellCast(enable)
  self.db.profile.spellcast = not not enable
  self:EnableHooksAndEvents()
end

function Core:SecureHookOrUnhook(enable, ...)
  local hooked = self:IsHooked((...))
  if enable then
    if not hooked then
      self:SecureHook(...)
    end
  elseif hooked then
    self:Unhook((...))
  end
end

function Core:EnableHooksAndEvents()
  if not self:IsActive() then return end
  self:SecureHookOrUnhook(self.db.profile.autorun, "ToggleAutoRun")

  self:SecureHookOrUnhook(self.db.profile.rightclick, "TurnOrActionStart")
  self:SecureHookOrUnhook(self.db.profile.rightclick, "TurnOrActionStop")

  local CURSOR_UPDATE = self:IsEventRegistered("CURSOR_UPDATE")
  if self.db.profile.spellcast then
    if not CURSOR_UPDATE then self:RegisterEvent("CURSOR_UPDATE") end
  elseif CURSOR_UPDATE then
    self:UnregisterEvent("CURSOR_UPDATE")
  end

  local AUTOFOLLOW_BEGIN = self:IsEventRegistered("AUTOFOLLOW_BEGIN")
  if self.db.profile.autofollow then
    if not AUTOFOLLOW_BEGIN then self:RegisterEvent("AUTOFOLLOW_BEGIN") end
  elseif AUTOFOLLOW_BEGIN then
    self:UnregisterEvent("AUTOFOLLOW_BEGIN")
  end

  SetMouselookOverrideBinding("BUTTON1", self.db.profile.leftclick_override or "NONE")
  SetMouselookOverrideBinding("BUTTON2", self.db.profile.rightclick_override or "NONE")
end

function Core:OnProfileEnable()
  self.db.profile.rightclick_delay = FixDelay(self.db.profile.rightclick_delay)

  self.db.profile.leftclick_override = FixClickOverride(self.db.profile.leftclick_override)
  self.db.profile.rightclick_override = FixClickOverride(self.db.profile.rightclick_override)

  self:EnableHooksAndEvents()
end

function Core:OnProfileDisable()
  self:MouselookOff()
  SetMouselookOverrideBinding("BUTTON1", nil)
  SetMouselookOverrideBinding("BUTTON2", nil)
end

function Core:ToggleMouselook()
  if IsMouselooking() then MouselookStop() else MouselookStart() end
end

function Core:MouselookOn()
  if not IsMouselooking() then MouselookStart() end
end

function Core:MouselookOff()
  if IsMouselooking() then MouselookStop() end
end

function Core:ToggleAutoRun()
  if SE_Movement:PlayerMoving() then return end
  self:MouselookOn()
end


do
  local time

  function Core:TurnOrActionStart()
    time = GetTime()
  end

  function Core:TurnOrActionStop()
    if not time then return end
    local dt = GetTime() - time
    time = nil
    if not IsMouselooking() and dt >= self:GetRightClickDelay() then
      MouselookStart()
    end
  end
end


function Core:CURSOR_UPDATE()
  if IsMouselooking() and SpellIsTargeting() then
    MouselookStop()
  end
end

function Core:AUTOFOLLOW_BEGIN()
  self:MouselookOff()
end


do
  local args = {}
  options = {
    type = "group",
    args = args,
  }

  args.autorun = {
    order = 100,
    type = "toggle",
    name = L["Auto-run"],
    desc = L["Enter mouse-look mode when you begin auto-running."],
    get = "IsAutoRun",
    set = "SetAutoRun",
  }

  args.autofollow = {
    order = 200,
    type = "toggle",
    name = L["Auto-follow"],
    desc = L["Leave mouse-look mode when you begin auto-following a unit."],
    get = "IsAutoFollow",
    set = "SetAutoFollow",
  }

  args.rightclick = {
    order = 300,
    type = "toggle",
    name = L["Right-click"],
    desc = L["Enter mouse-look mode when holding down the right mouse button for a short amount of time."],
    get = "IsRightClick",
    set = "SetRightClick",
  }
  args.rightclick_delay = {
    order = 310,
    type = "range",
    name = L["Right-click delay"],
    desc = L["Set the amount of time you need to hold down the right mouse button before entering mouse-look mode."],
    min = MinRightClickDelay,
    max = MaxRightClickDelay,
    get = "GetRightClickDelay",
    set = "SetRightClickDelay",
  }

  args.leftclick_override = {
    order = 400,
    type = "text",
    name = L["Left-click override"],
    desc = L["Adjust behaviour of the left mouse-button when you are in mouse-look mode."],
    get = "GetLeftClickOverride",
    set = "SetLeftClickOverride",
    validate = ClickOverrides,
  }
  args.rightclick_override = {
    order = 410,
    type = "text",
    name = L["Right-click override"],
    desc = L["Adjust behaviour of the right mouse-button when you are in mouse-look mode."],
    get = "GetRightClickOverride",
    set = "SetRightClickOverride",
    validate = ClickOverrides,
  }

  args.spellcast = {
    order = 500,
    type = "toggle",
    name = L["Spell-cast"],
    desc = L["Leave mouse-look mode when you cast a spell that needs a mouse-click to target."],
    get = "IsSpellCast",
    set = "SetSpellCast",
  }

  args.on = {
    guiHidden = true,
    type = "execute",
    name = L["Turn mouse-look on"],
    desc = L["Turn mouse-look on"],
    func = "MouselookOn",
  }

  args.off = {
    guiHidden = true,
    type = "execute",
    name = L["Turn mouse-look off"],
    desc = L["Turn mouse-look off"],
    func = "MouselookOff",
  }

  args.toggle = {
    guiHidden = true,
    type = "toggle",
    name = L["Toggle mouse-look on/off"],
    desc = L["Toggle mouse-look on/off"],
    get = function() return IsMouselooking() end,
    set = function(b) if b then Core:MouselookOn() else Core:MouselookOff() end end,
  }

  -- FuBar options
  Core.OnMenuRequest = options
  Core.hasIcon = "Interface\\Icons\\Ability_Hunter_AspectoftheViper"
  Core.hasNoColor = true
  Core.cannotDetachTooltip = true
  Core.independentProfile = true
  Core.hideWithoutStandby = true
end

_G.MouseLook = Core





